
#include "shared.h"


/* SMS context */
t_sms sms;


/* Run the virtual console emulation for one frame */
void sms_frame(int skip_render)
{
    /* Write frame marker to log file */
    if(snd.log) snd.callback(0x00);

    /* Take care of hard resets */
    if(input.system & INPUT_HARD_RESET)
    {
        system_reset();
    }

    /* Debounce pause key */
    if(input.system & INPUT_PAUSE)
    {
        if(!sms.paused)
        {
            sms.paused = 1;

            z80_set_nmi_line(ASSERT_LINE);
            z80_set_nmi_line(CLEAR_LINE);
        }
    }
    else
    {
         sms.paused = 0;
    }

    /* Run emulation for a single frame */
    for(vdp.line = 0; vdp.line < LINES_PER_FRAME; vdp.line += 1)
    {
        /* Handle VDP line events */
        vdp_run();

        /* Draw the current frame */
        if(!skip_render) render_line(vdp.line);

        /* Run the Z80 for a line */
        z80_execute(CYCLES_PER_LINE);
    }

    /* Update the emulated sound stream */
    if(snd.enabled) sound_update();
}


void sms_init(void)
{
#if PSX
    z80_set_context((void *)0x1F800000);
#endif

    z80_reset(0);
    z80_set_irq_callback(irq_callback);


    sms_reset();
}


void sms_reset(void)
{
    int save[2];

    /* Clear SMS context - ignore hack to preserve FM and country switches */
    save[0] = sms.country;
    save[1] = sms.fm_enable;
    memset(&sms, 0, sizeof(t_sms));
    sms.country = save[0];
    sms.fm_enable = save[1];

    sms.fcr[0] = 0x00;
    sms.fcr[1] = 0x00;
    sms.fcr[2] = 0x01;
    sms.fcr[3] = 0x00;

    /* Load memory maps with default values */
    cpu_readmap[0] = cart.rom + 0x0000; /* 0000-3FFF */
    cpu_readmap[1] = cart.rom + 0x2000;
    cpu_readmap[2] = cart.rom + 0x4000; /* 4000-7FFF */
    cpu_readmap[3] = cart.rom + 0x6000;
    cpu_readmap[4] = cart.rom + 0x0000; /* 0000-3FFF */
    cpu_readmap[5] = cart.rom + 0x2000;
    cpu_readmap[6] = sms.ram;            
    cpu_readmap[7] = sms.ram;

    cpu_writemap[0] = sms.dummy;         
    cpu_writemap[1] = sms.dummy;
    cpu_writemap[2] = sms.dummy;         
    cpu_writemap[3] = sms.dummy;
    cpu_writemap[4] = sms.dummy;         
    cpu_writemap[5] = sms.dummy;
    cpu_writemap[6] = sms.ram;           
    cpu_writemap[7] = sms.ram;

    cpu_readport  = (cart.type == TYPE_GG) ? gg_port_r : sms_port_r;
    cpu_writeport = (cart.type == TYPE_GG) ? gg_port_w : sms_port_w;

    sms.port_06 = 0xFF;
}

/* Write to memory */
void cpu_writemem16(int address, int data)
{
    cpu_writemap[(address >> 13)][(address & 0x1FFF)] = data;
    if(address >= 0xFFFC) sms_mapper_w(address & 3, data);
}


void gg_port_w(int port, int data)
{
    switch(port & 0xFF)
    {
        case 0x01: /* GG SIO */
        case 0x02:
        case 0x03:
        case 0x04:
        case 0x05:
            break;

        case 0x06: /* GG STEREO */
            if(snd.log)
            {
                snd.callback(0x04);
                snd.callback(data);
            }
            sms.port_06 = (data & 0xFF);
            break;

        case 0x7E: /* SN76489 PSG */
        case 0x7F:
            if(snd.log)
            {
                snd.callback(0x03);
                snd.callback(data);
            }
            if(snd.enabled) SN76496Write(0, data);
            break;

        case 0xBE: /* VDP DATA */
            vdp_data_w(data);
            break;

        case 0xBD: /* VDP CTRL */ 
        case 0xBF:
            vdp_ctrl_w(data);
            break;

        default:
            break;
    }
}


void sms_port_w(int port, int data)
{
    switch(port & 0xFF)
    {
        case 0x7E: /* SN76489 PSG */
        case 0x7F:
            if(snd.log)
            {
                snd.callback(0x03);
                snd.callback(data);
            }
            if(snd.enabled) SN76496Write(0, data);
            break;

        case 0xBE: /* VDP DATA */
            vdp_data_w(data);
            break;

        case 0xBD: /* VDP CTRL */
        case 0xBF:
            vdp_ctrl_w(data);
            break;

        case 0xF0: /* YM2413 */
        case 0xF1:
            if(snd.log)
            {
                snd.callback((port & 1) ? 0x06 : 0x05);
                snd.callback(data);
            }
            if(snd.enabled) ym2413_write(0, port & 1, data);
            break;

        case 0xF2: /* YM2413 DETECT */
            sms.port_F2 = (data & 1);
            break;

        case 0x3F: /* TERRITORY CTRL. */
            sms.port_3F = ((data & 0x80) | (data & 0x20) << 1) & 0xC0;
            if(sms.country == TYPE_DOMESTIC) sms.port_3F ^= 0xC0;
            break;

        default:
            break;
    }
}


int gg_port_r(int port)
{
    byte temp = 0xFF;

    switch(port & 0xFF)
    {
        case 0x01: /* GG SIO */
        case 0x02:
        case 0x03:
        case 0x04:
        case 0x05:
            return (0x00);
    
        case 0x7E: /* V COUNTER */
            return (vdp_vcounter_r());
            break;
    
        case 0x7F: /* H COUNTER */
            return (vdp_hcounter_r());
            break;
    
        case 0x00: /* INPUT #2 */
            temp = 0xFF;
            if(input.system & INPUT_START) temp &= ~0x80;
            if(sms.country == TYPE_DOMESTIC) temp &= ~0x40;
            return (temp);
    
        case 0xC0: /* INPUT #0 */  
        case 0xDC:
            temp = 0xFF;
            if(input.pad[0] & INPUT_UP)      temp &= ~0x01;
            if(input.pad[0] & INPUT_DOWN)    temp &= ~0x02;
            if(input.pad[0] & INPUT_LEFT)    temp &= ~0x04;
            if(input.pad[0] & INPUT_RIGHT)   temp &= ~0x08;
            if(input.pad[0] & INPUT_BUTTON2) temp &= ~0x10;
            if(input.pad[0] & INPUT_BUTTON1) temp &= ~0x20;
            return (temp);
    
        case 0xBE: /* VDP DATA */
            return (vdp_data_r());
  
        case 0xBD:
        case 0xBF: /* VDP CTRL */
            return (vdp_ctrl_r());

        default:
            break;
    }
    return (temp);
}


int sms_port_r(int port)
{
    byte temp = 0xFF;

    switch(port & 0xFF)
    {
        case 0x7E: /* V COUNTER */
            return (vdp_vcounter_r());
            break;
    
        case 0x7F: /* H COUNTER */
            return (vdp_hcounter_r());
            break;
    
        case 0xC0: /* INPUT #0 */  
        case 0xDC:
            temp = 0xFF;
            if(input.pad[0] & INPUT_UP)      temp &= ~0x01;
            if(input.pad[0] & INPUT_DOWN)    temp &= ~0x02;
            if(input.pad[0] & INPUT_LEFT)    temp &= ~0x04;
            if(input.pad[0] & INPUT_RIGHT)   temp &= ~0x08;
            if(input.pad[0] & INPUT_BUTTON2) temp &= ~0x10;
            if(input.pad[0] & INPUT_BUTTON1) temp &= ~0x20;
            if(input.pad[1] & INPUT_UP)      temp &= ~0x40;
            if(input.pad[1] & INPUT_DOWN)    temp &= ~0x80;
            return (temp);
    
        case 0xC1: /* INPUT #1 */
        case 0xDD:
            temp = 0xFF;
            if(input.pad[1] & INPUT_LEFT)       temp &= ~0x01;
            if(input.pad[1] & INPUT_RIGHT)      temp &= ~0x02;
            if(input.pad[1] & INPUT_BUTTON2)    temp &= ~0x04;
            if(input.pad[1] & INPUT_BUTTON1)    temp &= ~0x08;
            if(input.system & INPUT_SOFT_RESET) temp &= ~0x10;
            return ((temp & 0x3F) | (sms.port_3F & 0xC0));
    
        case 0xBE: /* VDP DATA */
            return (vdp_data_r());
   
        case 0xBD:
        case 0xBF: /* VDP CTRL */
            return (vdp_ctrl_r());
    
        case 0xF2: /* YM2413 DETECT */
            if(sms.fm_enable) return (sms.port_F2);
            break;

        default:
            break;
    }
    return (temp);
}


void sms_mapper_w(int address, int data)
{
    /* Calculate ROM page index */
    byte page = (data % cart.pages);

    /* Save frame control register data */
    sms.fcr[address] = data;

    switch(address)
    {
        case 0:
            if(data & 8)
            {
                sms.save = 1;

                /* Page in RAM */
                cpu_readmap[4]  = &sms.sram[(data & 4) ? 0x4000 : 0x0000];
                cpu_readmap[5]  = &sms.sram[(data & 4) ? 0x6000 : 0x2000];
                cpu_writemap[4] = &sms.sram[(data & 4) ? 0x4000 : 0x0000];
                cpu_writemap[5] = &sms.sram[(data & 4) ? 0x6000 : 0x2000];
            }
            else
            {
                /* Page in ROM */
                cpu_readmap[4]  = &cart.rom[((sms.fcr[3] % cart.pages) << 14) + 0x0000];
                cpu_readmap[5]  = &cart.rom[((sms.fcr[3] % cart.pages) << 14) + 0x2000];
                cpu_writemap[4] = sms.dummy;
                cpu_writemap[5] = sms.dummy;
            }
            break;

        case 1:
            cpu_readmap[0] = &cart.rom[(page << 14) + 0x0000];
            cpu_readmap[1] = &cart.rom[(page << 14) + 0x2000];
            break;

        case 2:
            cpu_readmap[2] = &cart.rom[(page << 14) + 0x0000];
            cpu_readmap[3] = &cart.rom[(page << 14) + 0x2000];
            break;

        case 3:
            if(!(sms.fcr[0] & 0x08))
            {
                cpu_readmap[4] = &cart.rom[(page << 14) + 0x0000];
                cpu_readmap[5] = &cart.rom[(page << 14) + 0x2000];
            }
            break;
    }
}

int irq_callback(int param)
{
    return (0xFF);
}


void sound_update(void)
{
    int count;
	int stereo_count;

    SN76496Update(0, snd.psg_buffer, snd.bufsize, sms.port_06);
    YM3812UpdateOne(ym3812, snd.fm_buffer, snd.bufsize);

    for(count = 0, stereo_count = 0 ; count < snd.bufsize; count += 1)
    {
        snd.buffer[stereo_count++] = snd.fm_buffer[count] + snd.psg_buffer[0][count];
		snd.buffer[stereo_count++] = snd.fm_buffer[count] + snd.psg_buffer[1][count];
    }
}


